@@ -1,11 +1,11 @@
 from __future__ import annotations
 
 import asyncio
+import functools
 import inspect
 import json
 import re
 import threading
-import time
 import weakref
 from abc import ABC, abstractmethod
 from collections import OrderedDict
@@ -20,7 +20,6 @@ from typing import (
     Iterator,
     List,
     Mapping,
-    NamedTuple,
     Optional,
     Protocol,
     Set,
@@ -33,16 +32,12 @@ from typing import (
     runtime_checkable,
 )
 
-from robotcode.core.async_tools import (
-    HasThreaded,
-    async_event,
-    create_sub_future,
-    create_sub_task,
-    run_coroutine_in_thread,
-)
-from robotcode.core.dataclasses import as_json, from_dict
-from robotcode.core.logging import LoggingDescriptor
+from robotcode.core.async_tools import run_coroutine_in_thread
+from robotcode.core.concurrent import Task, run_as_task
+from robotcode.core.event import event
+from robotcode.core.utils.dataclasses import as_json, from_dict
 from robotcode.core.utils.inspect import ensure_coroutine, iter_methods
+from robotcode.core.utils.logging import LoggingDescriptor
 
 __all__ = [
     "JsonRPCErrors",
@@ -145,11 +140,21 @@ class InvalidProtocolVersionError(JsonRPCParseError):
     pass
 
 
-class RpcMethodEntry(NamedTuple):
+@dataclass
+class RpcMethodEntry:
     name: str
     method: Callable[..., Any]
     param_type: Optional[Type[Any]]
     cancelable: bool
+    threaded: bool
+
+    _is_coroutine: Optional[bool] = field(default=None, init=False)
+
+    @property
+    def is_coroutine(self) -> bool:
+        if self._is_coroutine is None:
+            self._is_coroutine = inspect.iscoroutinefunction(self.method)
+        return self._is_coroutine
 
 
 @runtime_checkable
@@ -169,7 +174,11 @@ def rpc_method(_func: _F) -> _F:
 
 @overload
 def rpc_method(
-    *, name: Optional[str] = None, param_type: Optional[Type[Any]] = None, cancelable: bool = True
+    *,
+    name: Optional[str] = None,
+    param_type: Optional[Type[Any]] = None,
+    cancelable: bool = True,
+    threaded: bool = False,
 ) -> Callable[[_F], _F]:
     ...
 
@@ -180,6 +189,7 @@ def rpc_method(
     name: Optional[str] = None,
     param_type: Optional[Type[Any]] = None,
     cancelable: bool = True,
+    threaded: bool = False,
 ) -> Callable[[_F], _F]:
     def _decorator(func: _F) -> Callable[[_F], _F]:
         if inspect.isclass(_func):
@@ -196,7 +206,7 @@ def rpc_method(
         if real_name is None or not real_name:
             raise ValueError("name is empty.")
 
-        cast(RpcMethod, f).__rpc_method__ = RpcMethodEntry(real_name, f, param_type, cancelable)
+        cast(RpcMethod, f).__rpc_method__ = RpcMethodEntry(real_name, f, param_type, cancelable, threaded)
         return func
 
     if _func is None:
@@ -222,7 +232,7 @@ class RpcRegistry:
     def __set_name__(self, owner: Any, name: str) -> None:
         self.__owner = owner
 
-    def __get__(self, obj: Any, obj_type: Type[Any]) -> "RpcRegistry":
+    def __get__(self, obj: Any, obj_type: Type[Any]) -> RpcRegistry:
         if obj is None and obj_type == self.__owner:
             return self
 
@@ -263,6 +273,7 @@ class RpcRegistry:
                     method,
                     rpc_method.__rpc_method__.param_type,
                     rpc_method.__rpc_method__.cancelable,
+                    rpc_method.__rpc_method__.threaded,
                 )
                 for method, rpc_method in map(
                     lambda m1: (m1, cast(RpcMethod, m1)),
@@ -308,11 +319,16 @@ class RpcRegistry:
         return self.__methods
 
     def add_method(
-        self, name: str, func: Callable[..., Any], param_type: Optional[Type[Any]] = None, cancelable: bool = True
+        self,
+        name: str,
+        func: Callable[..., Any],
+        param_type: Optional[Type[Any]] = None,
+        cancelable: bool = True,
+        threaded: bool = False,
     ) -> None:
         self.__ensure_initialized()
 
-        self.__methods[name] = RpcMethodEntry(name, func, param_type, cancelable)
+        self.__methods[name] = RpcMethodEntry(name, func, param_type, cancelable, threaded)
 
     def remove_method(self, name: str) -> Optional[RpcMethodEntry]:
         self.__ensure_initialized()
@@ -335,15 +351,28 @@ class RpcRegistry:
         return result.param_type
 
 
-class SendedRequestEntry(NamedTuple):
-    future: asyncio.Future[Any]
-    result_type: Optional[Type[Any]]
+class SendedRequestEntry:
+    def __init__(self, future: Task[Any], result_type: Optional[Type[Any]]) -> None:
+        self.future = future
+        self.result_type = result_type
 
 
-class ReceivedRequestEntry(NamedTuple):
-    future: asyncio.Future[Any]
-    request: Optional[Any]
-    cancelable: bool
+class ReceivedRequestEntry:
+    def __init__(
+        self,
+        future: asyncio.Future[Any],
+        request: JsonRPCRequest,
+        cancelable: bool,
+    ) -> None:
+        self.future = future
+        self.request = request
+        self.cancelable = cancelable
+        self.cancel_requested = False
+
+    def cancel(self) -> None:
+        self.cancel_requested = True
+        if self.future is not None and not self.future.cancelled():
+            self.future.cancel()
 
 
 class JsonRPCProtocolBase(asyncio.Protocol, ABC):
@@ -352,19 +381,19 @@ class JsonRPCProtocolBase(asyncio.Protocol, ABC):
     def __init__(self) -> None:
         self.read_transport: Optional[asyncio.ReadTransport] = None
         self.write_transport: Optional[asyncio.WriteTransport] = None
-        self._message_buf = bytes()
+        self._message_buf = b""
         self._loop: Optional[asyncio.AbstractEventLoop] = None
 
     @property
     def loop(self) -> Optional[asyncio.AbstractEventLoop]:
         return self._loop
 
-    @async_event
-    async def on_connection_made(sender, transport: asyncio.BaseTransport) -> None:
+    @event
+    def on_connection_made(sender, transport: asyncio.BaseTransport) -> None:
         ...
 
-    @async_event
-    async def on_connection_lost(sender, exc: Optional[BaseException]) -> None:
+    @event
+    def on_connection_lost(sender, exc: Optional[BaseException]) -> None:
         ...
 
     def connection_made(self, transport: asyncio.BaseTransport) -> None:
@@ -375,10 +404,10 @@ class JsonRPCProtocolBase(asyncio.Protocol, ABC):
         if isinstance(transport, asyncio.WriteTransport):
             self.write_transport = transport
 
-        create_sub_task(self.on_connection_made(self, transport))
+        self.on_connection_made(self, transport)
 
     def connection_lost(self, exc: Optional[BaseException]) -> None:
-        create_sub_task(self.on_connection_lost(self, exc))
+        self.on_connection_lost(self, exc)
         self._loop = None
 
     def eof_received(self) -> Optional[bool]:
@@ -415,7 +444,7 @@ class JsonRPCProtocolBase(asyncio.Protocol, ABC):
                 return
 
             body, data = body[:length], body[length:]
-            self._message_buf = bytes()
+            self._message_buf = b""
 
             self._handle_body(body, charset)
 
@@ -435,10 +464,12 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         self._sended_request_count = 0
         self._received_request: OrderedDict[Union[str, int, None], ReceivedRequestEntry] = OrderedDict()
         self._received_request_lock = threading.RLock()
+        self._signature_cache: Dict[Callable[..., Any], inspect.Signature] = {}
+        self._running_handle_message_tasks: Set[asyncio.Future[Any]] = set()
 
     @staticmethod
     def _generate_json_rpc_messages_from_dict(
-        data: Union[Dict[Any, Any], List[Dict[Any, Any]]]
+        data: Union[Dict[Any, Any], List[Dict[Any, Any]]],
     ) -> Iterator[JsonRPCMessage]:
         def inner(d: Dict[Any, Any]) -> JsonRPCMessage:
             if "jsonrpc" in d:
@@ -478,15 +509,10 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
             self.send_error(JsonRPCErrors.PARSE_ERROR, f"{type(e).__name__}: {e}")
 
     def _handle_messages(self, iterator: Iterator[JsonRPCMessage]) -> None:
-        def done(f: asyncio.Future[Any]) -> None:
-            if f.done() and not f.cancelled():
-                ex = f.exception()
-
-                if ex is None or isinstance(ex, asyncio.CancelledError):
-                    return
-
         for m in iterator:
-            create_sub_task(self.handle_message(m)).add_done_callback(done)
+            task = asyncio.create_task(self.handle_message(m))
+            self._running_handle_message_tasks.add(task)
+            task.add_done_callback(self._running_handle_message_tasks.discard)
 
     @__logger.call
     async def handle_message(self, message: JsonRPCMessage) -> None:
@@ -515,12 +541,7 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         if data is not None:
             error_obj.data = data
 
-        self.send_message(
-            JsonRPCError(
-                id=id,
-                error=error_obj,
-            )
-        )
+        self.send_message(JsonRPCError(id=id, error=error_obj))
 
     @__logger.call
     def send_message(self, message: JsonRPCMessage) -> None:
@@ -545,27 +566,28 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         self,
         method: str,
         params: Optional[Any] = None,
-        return_type_or_converter: Optional[Type[_TResult]] = None,
-    ) -> asyncio.Future[Optional[_TResult]]:
+        return_type: Optional[Type[_TResult]] = None,
+    ) -> Task[_TResult]:
+        result: Task[_TResult] = Task()
+
         with self._sended_request_lock:
-            result: asyncio.Future[Optional[_TResult]] = create_sub_future()
             self._sended_request_count += 1
             id = self._sended_request_count
 
-            self._sended_request[id] = SendedRequestEntry(result, return_type_or_converter)
+            self._sended_request[id] = SendedRequestEntry(result, return_type)
 
         request = JsonRPCRequest(id=id, method=method, params=params)
         self.send_message(request)
 
         return result
 
-    async def send_request_async(
+    def send_request_async(
         self,
         method: str,
         params: Optional[Any] = None,
-        return_type: Optional[Type[_TResult]] = None,
-    ) -> Optional[_TResult]:
-        return await self.send_request(method, params, return_type)
+        return_type_or_converter: Optional[Type[_TResult]] = None,
+    ) -> asyncio.Future[_TResult]:
+        return asyncio.wrap_future(self.send_request(method, params, return_type_or_converter))
 
     @__logger.call
     def send_notification(self, method: str, params: Any) -> None:
@@ -590,44 +612,17 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
 
         try:
             if not entry.future.done():
-                res = None
-                if message.result is not None:
-                    res = from_dict(message.result, entry.result_type)
-                if entry.future.get_loop() == asyncio.get_running_loop():
-                    entry.future.set_result(res)
-                else:
-                    if entry.future.get_loop().is_running():
-
-                        def set_result(f: asyncio.Future[Any], r: Any, ev: threading.Event) -> None:
-                            try:
-                                if not f.done() and f.get_loop().is_running():
-                                    f.set_result(r)
-                            finally:
-                                ev.set()
-
-                        done = threading.Event()
-
-                        entry.future.get_loop().call_soon_threadsafe(set_result, entry.future, res, done)
-
-                        start = time.monotonic()
-                        while not done.is_set():
-                            if time.monotonic() - start > 120:
-                                raise TimeoutError("Can't set future result.")
-
-                            await asyncio.sleep(0)
-
-                    else:
-                        self.__logger.warning(lambda: f"Response {entry!r} loop is not running.")
+                entry.future.set_result(
+                    from_dict(message.result, entry.result_type) if message.result is not None else None
+                )
+            else:
+                self.__logger.warning(lambda: f"Response for {message} is already done.")
 
         except (SystemExit, KeyboardInterrupt):
             raise
         except BaseException as e:
             if not entry.future.done():
-                if entry.future.get_loop() == asyncio.get_running_loop():
-                    entry.future.set_exception(e)
-                else:
-                    if entry.future.get_loop().is_running():
-                        entry.future.get_loop().call_soon_threadsafe(entry.future.set_exception, e)
+                entry.future.set_exception(e)
 
     @__logger.call
     async def handle_error(self, message: JsonRPCError) -> None:
@@ -646,54 +641,27 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
 
         try:
             if not entry.future.done():
-                res = None
-                if message.result is not None:
-                    res = from_dict(message.result, entry.result_type)
-                if entry.future.get_loop() == asyncio.get_running_loop():
-                    entry.future.set_exception(
-                        JsonRPCErrorException(message.error.code, message.error.message, message.error.data)
+                entry.future.set_exception(
+                    JsonRPCErrorException(
+                        message.error.code,
+                        message.error.message,
+                        message.error.data,
                     )
-                else:
-                    if entry.future.get_loop().is_running():
-
-                        def set_result(f: asyncio.Future[Any], r: Any, ev: threading.Event) -> None:
-                            try:
-                                if not f.done() and f.get_loop().is_running():
-                                    f.set_exception(
-                                        JsonRPCErrorException(
-                                            message.error.code, message.error.message, message.error.data
-                                        )
-                                    )
-                            finally:
-                                ev.set()
-
-                        done = threading.Event()
-
-                        entry.future.get_loop().call_soon_threadsafe(set_result, entry.future, res, done)
-
-                        start = time.monotonic()
-                        while not done.is_set():
-                            if time.monotonic() - start > 120:
-                                raise TimeoutError("Can't set future result.")
-
-                            await asyncio.sleep(0)
-
-                    else:
-                        self.__logger.warning(lambda: f"Response {entry!r} loop is not running.")
+                )
+            else:
+                self.__logger.warning(lambda: f"Error Response for {message} is already done.")
 
         except (SystemExit, KeyboardInterrupt):
             raise
         except BaseException as e:
             if not entry.future.done():
-                if entry.future.get_loop() == asyncio.get_running_loop():
-                    entry.future.set_exception(e)
-                else:
-                    if entry.future.get_loop().is_running():
-                        entry.future.get_loop().call_soon_threadsafe(entry.future.set_exception, e)
+                entry.future.set_exception(e)
 
-    @staticmethod
     def _convert_params(
-        callable: Callable[..., Any], params_type: Optional[Type[Any]], params: Any
+        self,
+        callable: Callable[..., Any],
+        params_type: Optional[Type[Any]],
+        params: Any,
     ) -> Tuple[List[Any], Dict[str, Any]]:
         if params is None:
             return [], {}
@@ -706,7 +674,12 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         # try to convert the dict to correct type
         converted_params = from_dict(params, params_type)
 
-        signature = inspect.signature(callable)
+        # get the signature of the callable
+        if callable in self._signature_cache:
+            signature = self._signature_cache[callable]
+        else:
+            signature = inspect.signature(callable)
+            self._signature_cache[callable] = signature
 
         has_var_kw = any(p.kind == inspect.Parameter.VAR_KEYWORD for p in signature.parameters.values())
 
@@ -756,78 +729,107 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         return args, kw_args
 
     async def handle_request(self, message: JsonRPCRequest) -> None:
-        e = self.registry.get_entry(message.method)
+        try:
+            e = self.registry.get_entry(message.method)
 
-        if e is None or not callable(e.method):
-            self.send_error(
-                JsonRPCErrors.METHOD_NOT_FOUND,
-                f"Unknown method: {message.method}",
-                id=message.id,
-            )
-            return
+            if e is None or not callable(e.method):
+                self.send_error(
+                    JsonRPCErrors.METHOD_NOT_FOUND,
+                    f"Unknown method: {message.method}",
+                    id=message.id,
+                )
+                return
 
-        params = self._convert_params(e.method, e.param_type, message.params)
+            params = self._convert_params(e.method, e.param_type, message.params)
 
-        if (
-            isinstance(e.method, HasThreaded)
-            and e.method.__threaded__
-            or inspect.ismethod(e.method)
-            and isinstance(e.method.__func__, HasThreaded)
-            and e.method.__func__.__threaded__
-        ):
-            task = run_coroutine_in_thread(
-                ensure_coroutine(cast(Callable[..., Any], e.method)), *params[0], **params[1]
-            )
-        else:
-            task = create_sub_task(
-                ensure_coroutine(e.method)(*params[0], **params[1]),
-                name=message.method,
-            )
+            if not e.threaded and not e.is_coroutine:
+                self.send_response(message.id, e.method(*params[0], **params[1]))
+            else:
+                if e.threaded:
+                    if e.is_coroutine:
+                        task = run_coroutine_in_thread(
+                            ensure_coroutine(cast(Callable[..., Any], e.method)),
+                            *params[0],
+                            **params[1],
+                        )
+                    else:
+                        task = asyncio.wrap_future(run_as_task(e.method, *params[0], **params[1]))
+                else:
+                    task = asyncio.create_task(e.method(*params[0], **params[1]), name=message.method)
 
-        with self._received_request_lock:
-            self._received_request[message.id] = ReceivedRequestEntry(task, message, e.cancelable)
+                with self._received_request_lock:
+                    self._received_request[message.id] = ReceivedRequestEntry(task, message, e.cancelable)
 
-        def done(t: asyncio.Future[Any]) -> None:
-            try:
+                task.add_done_callback(functools.partial(self._received_request_done, message))
+
+                await task
+        except (SystemExit, KeyboardInterrupt, asyncio.CancelledError):
+            raise
+        except BaseException as e:
+            self.__logger.exception(e)
+
+    def _received_request_done(self, message: JsonRPCRequest, t: asyncio.Future[Any]) -> None:
+        try:
+            with self._received_request_lock:
+                entry = self._received_request.pop(message.id, None)
+
+            if entry is None:
+                self.__logger.critical(lambda: f"unknown request {message!r}")
+                return
+
+            if entry.cancel_requested:
+                self.__logger.debug(lambda: f"request {message!r} canceled")
+                self.send_error(
+                    JsonRPCErrors.REQUEST_CANCELLED,
+                    "Request canceled.",
+                    id=message.id,
+                )
+            else:
                 if not t.cancelled():
                     ex = t.exception()
                     if ex is not None:
                         self.__logger.exception(ex, exc_info=ex)
-                        raise JsonRPCErrorException(JsonRPCErrors.INTERNAL_ERROR, f"{type(ex).__name__}: {ex}") from ex
+                        raise JsonRPCErrorException(
+                            JsonRPCErrors.INTERNAL_ERROR,
+                            f"{type(ex).__name__}: {ex}",
+                        ) from ex
 
                 self.send_response(message.id, t.result())
-            except asyncio.CancelledError:
-                self.__logger.debug(lambda: f"request message {message!r} canceled")
-                self.send_error(JsonRPCErrors.REQUEST_CANCELLED, "Request canceled.", id=message.id)
-            except (SystemExit, KeyboardInterrupt):
-                raise
-            except JsonRPCErrorException as e:
-                self.send_error(e.code, e.message or f"{type(e).__name__}: {e}", id=message.id, data=e.data)
-            except BaseException as e:
-                self.__logger.exception(e)
-                self.send_error(JsonRPCErrors.INTERNAL_ERROR, f"{type(e).__name__}: {e}", id=message.id)
-            finally:
-                with self._received_request_lock:
-                    self._received_request.pop(message.id, None)
+        except asyncio.CancelledError:
+            self.__logger.debug(lambda: f"request message {message!r} canceled")
+            self.send_error(
+                JsonRPCErrors.REQUEST_CANCELLED,
+                "Request canceled.",
+                id=message.id,
+            )
+        except (SystemExit, KeyboardInterrupt):
+            raise
+        except JsonRPCErrorException as e:
+            self.send_error(
+                e.code,
+                e.message or f"{type(e).__name__}: {e}",
+                id=message.id,
+                data=e.data,
+            )
+        except BaseException as e:
+            self.__logger.exception(e)
+            self.send_error(
+                JsonRPCErrors.INTERNAL_ERROR,
+                f"{type(e).__name__}: {e}",
+                id=message.id,
+            )
 
-        task.add_done_callback(done)
-
-        await task
-
-    @__logger.call
     def cancel_request(self, id: Union[int, str, None]) -> None:
         with self._received_request_lock:
             entry = self._received_request.get(id, None)
 
-        if entry is not None and entry.future is not None and not entry.future.cancelled():
+        if entry is not None:
             self.__logger.debug(lambda: f"try to cancel request {entry.request if entry is not None else ''}")
-            entry.future.cancel()
+            entry.cancel()
 
-    @__logger.call
-    async def cancel_all_received_request(self) -> None:
+    def cancel_all_received_request(self) -> None:
         for entry in self._received_request.values():
-            if entry is not None and entry.cancelable and entry.future is not None and not entry.future.cancelled():
-                entry.future.cancel()
+            entry.cancel()
 
     @__logger.call
     async def handle_notification(self, message: JsonRPCNotification) -> None:
@@ -839,23 +841,20 @@ class JsonRPCProtocol(JsonRPCProtocolBase):
         try:
             params = self._convert_params(e.method, e.param_type, message.params)
 
-            if (
-                isinstance(e.method, HasThreaded)
-                and e.method.__threaded__
-                or inspect.ismethod(e.method)
-                and isinstance(e.method.__func__, HasThreaded)
-                and e.method.__func__.__threaded__
-            ):
-                task = run_coroutine_in_thread(
-                    ensure_coroutine(cast(Callable[..., Any], e.method)), *params[0], **params[1]
-                )
+            if not e.threaded and not e.is_coroutine:
+                e.method(*params[0], **params[1])
             else:
-                task = create_sub_task(
-                    ensure_coroutine(e.method)(*params[0], **params[1]),
-                    name=message.method,
-                )
+                if e.threaded:
+                    if e.is_coroutine:
+                        task = run_coroutine_in_thread(
+                            ensure_coroutine(cast(Callable[..., Any], e.method)), *params[0], **params[1]
+                        )
+                    else:
+                        task = asyncio.wrap_future(run_as_task(e.method, *params[0], **params[1]))
+                else:
+                    task = asyncio.create_task(e.method(*params[0], **params[1]), name=message.method)
 
-            await task
+                await task
         except asyncio.CancelledError:
             pass
         except (SystemExit, KeyboardInterrupt):
@@ -869,15 +868,12 @@ TProtocol = TypeVar("TProtocol", bound=JsonRPCProtocol)
 
 class GenericJsonRPCProtocolPart(Generic[TProtocol]):
     def __init__(self, parent: TProtocol) -> None:
-        self._parent = weakref.ref(parent)
+        self._parent = parent
         parent.registry.add_class_part_instance(self)
 
     @property
     def parent(self) -> TProtocol:
-        result = self._parent()
-        if result is None:
-            raise JsonRPCException("WeakRef is dead.")
-        return result
+        return self._parent
 
 
 class JsonRPCProtocolPart(GenericJsonRPCProtocolPart[JsonRPCProtocol]):
