@@ -18,10 +18,11 @@ from typing import (
     cast,
 )
 
-from robotcode.core.logging import LoggingDescriptor
-from robotcode.core.types import ServerMode, TcpParams
 from typing_extensions import Self
 
+from robotcode.core.types import ServerMode, TcpParams
+from robotcode.core.utils.logging import LoggingDescriptor
+
 from .protocol import JsonRPCException
 
 __all__ = ["JsonRPCServer"]
@@ -225,13 +226,16 @@ class JsonRPCServer(Generic[TProtocol], abc.ABC):
             #  create_pipe_connection method
             if sys.platform == "win32" and hasattr(self.loop, "create_pipe_connection"):
                 if TYPE_CHECKING:
-                    from asyncio.streams import StreamReaderProtocol
                     from asyncio.windows_events import ProactorEventLoop
 
                 self.loop.run_until_complete(
                     cast("ProactorEventLoop", self.loop).create_pipe_connection(
-                        lambda: cast("StreamReaderProtocol", self.create_protocol()), pipe_name
-                    ),
+                        lambda: cast(
+                            "asyncio.StreamReaderProtocol",
+                            self.create_protocol(),
+                        ),
+                        pipe_name,
+                    )
                 )
             else:
                 self.loop.run_until_complete(self.loop.create_unix_connection(self.create_protocol, pipe_name))
@@ -254,11 +258,11 @@ class JsonRPCServer(Generic[TProtocol], abc.ABC):
             #  create_pipe_connection method
             if sys.platform == "win32" and hasattr(self.loop, "start_serving_pipe"):
                 if TYPE_CHECKING:
-                    from asyncio.streams import StreamReaderProtocol
                     from asyncio.windows_events import ProactorEventLoop
 
                 await cast("ProactorEventLoop", self.loop).start_serving_pipe(
-                    lambda: cast("StreamReaderProtocol", self.create_protocol()), pipe_name
+                    lambda: cast("asyncio.StreamReaderProtocol", self.create_protocol()),
+                    pipe_name,
                 )
 
             else:
